==> tuxurf-0.02.tar.gz: HTW/token.c <==
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "parser.h"


struct token_stack *tokens = NULL;
struct token_stack *tokens_tail = NULL;
struct token_stack *stack = NULL;

static void token_push(struct token_stack *ts)
{
	ts->link = NULL;
	if(tokens == NULL)
		tokens = ts;
	else
		tokens_tail->link = ts;
	tokens_tail = ts;
	ts->next = stack;
	if(stack && stack->done == 0)
	{
		render_text(stack->data);
		stack->done = 1;
	}
	ts->render = find_renderer(ts->tag);
	if(ts->render)
		ts->render->begin(ts);
	stack = ts;
}

static void token_pop_to(char *tag)
{
	char buf[256];
	struct token_stack *ts = stack;

	while(ts!=NULL)
	{
		if(strcasecmp(ts->tag, tag)==0)
			break;
		ts=ts->next;
	}	
	if(ts==NULL)
	{
		snprintf(buf, 256, "No opening tag found for '%s'", tag);
		parser_warning(buf);
		return;
	}
	/*
	 *	Pop entries
	 */
	
	if(!stack->done)
		render_text(stack->data);
	stack->done = 1;
	
	while(stack != ts->next)
	{
		if(stack->render)
			stack->render->end(stack);
		stack = stack->next;
	}
}

/*
 *	Handle implied closures
 */
 
void token_peek_pop(char *tag, char *until)
{
	if(stack && strcasecmp(stack->tag, tag)==0)
	{
		token_pop_to(tag);
		return;
	}
	if(until)
	{
		struct token_stack *p = stack;
		while(p!=NULL)
		{
			/* Found the bounding tag first. No implies */
			if(strcasecmp(p->tag, until)==0)
				return;
			/* Found implied close, pop back to it */
			if(strcasecmp(p->tag, tag)==0)
			{
				token_pop_to(tag);
				return;
			}
			p=p->next;
		}
	}
}
	
static void tag_append(struct token_stack *token, struct token_attr *attr)
{
	struct token_attr **a = &token->attr;
	
	while((*a)!=NULL)
	{
		if(strcasecmp((*a)->key, attr->key)==0)
			parser_warning("Duplicate attribute");
		a=&((*a)->next);
	}
	attr->next = NULL;
	*a = attr;
}

char *tag_attr_default(struct token_stack *token, char *key, char *def)
{
	struct token_attr *a = token->attr;
	
	while(a!=NULL)
	{
		if(strcasecmp(a->key, key)==0)
			return a->val;
		a=a->next;
	}
	return def;
}
	
char *tag_attr(struct token_stack *token, char *key)
{
	return tag_attr_default(token, key, NULL);
}

static struct token_stack *tag_alloc(char *key)
{
	struct token_stack *token = (struct token_stack *)malloc(sizeof(struct token_stack) + strlen(key)+1);
	if(token==NULL)
		parser_fatal("Out of memory");
	token->tag = (char *)(token+1);
	strcpy(token->tag, key);
	token->next = NULL;
	token->link = NULL;
	token->done = 0;
	token->render = NULL;
	token->data = NULL;
	token->rval = 0L;
	token->attr = NULL;
	return token;
}

static struct token_attr *tag_attr_alloc(char * key, char *val)
{
	struct token_attr *attr;
	
	
	attr = (struct token_attr *)malloc(sizeof(struct token_attr) +
						strlen(key)+strlen(val)+2);
	if(attr==NULL)
		parser_fatal("Out of memory");
	attr->key = (char *)(attr+1);
	attr->val = attr->key + strlen(key)+1;
	attr->next = NULL;
	strcpy(attr->key, key);
	strcpy(attr->val, val);
	return attr;
}

static char *parse_quoted_value(char *p, char **v, int *done)
{
	char quote = *p++;

	*v = p;
		
	while(*p!=0 && *p!=quote)
	{
		p++;
	}
	if(*p==0)
	{
		parser_warning("Missing quote");
		*done=1;
		return p;
	}
	else
		*p++=0;
		
	return p;
}	
	
static char *parse_attr_value(char *p, char **v, int *done)
{
	static char nullstr[]="";
	
	if(*p==0 || *p=='>')
	{
		*v=nullstr;
		return p;
	}
	
	while(p && isspace(*p))
		p++;
		
	if(*p==0)
		return p;
		
	if(*p=='\'' || *p=='"')
		return parse_quoted_value(p, v, done);
		
	*v = p;
	while(*p && !isspace(*p) && *p!='>')
		p++;
	if(isspace(*p))
		*p++=0;
	if(*p=='>')
	{
		*p++=0;
		*done=1;
	}
	return p;
}

static char *parse_attr(struct token_stack *token, char *p, int *done)
{
	struct token_attr *attr;
	char *k;
	char *v = "";
	while(*p && isspace(*p))
		p++;
	if(*p==0 || *p=='>')
	{
		*done = 1;
		if(*p=='>')
			*p++=0;
		return p;
	}
	k=p;
	while(*p && !isspace(*p) && *p != '>' && *p != '=')
		p++;
	if(*p=='=')
	{
		*p++=0;
		p = parse_attr_value(p, &v, done);
	}
	else if(*p == '>')
	{
		*done = 1;
		*p++=0;
	}
	else if(*p)
		*p++=0;
				
	attr = tag_attr_alloc(k, v);
	tag_append(token, attr);
	return p;		
}


static char *parse_tag(char *p)
{
	char *tb = p;
	int done = 0;
	struct token_stack *token;
	char buf[32];
	
	/*
	 *	Check for SGML comment [CDATA]
	 */
	 
	if(memcmp(p, "!--", 3)==0)
	{
		while(*p && strncmp(p, "-->",3))
			p++;
		return p;
	}
	
	while(*p)
	{
		if(isspace(*p))
			break;
		if(*p=='>')
			break;
		p++;
	}
	if(*p==0)
	{
		parser_warning("Unexpected end of document");
		return p;
	}
	if(*p=='>')
		done = 1;
		
	*p++ = 0;
	
	token = tag_alloc(tb);
	
	while(*p && *p!='>' && !done)
		p=parse_attr(token, p, &done);
	
	if(*p==0 && !done)
		parser_warning("Unexpected end of document");
	
	if(*p=='>')
		*p++ = 0;
		
	/*
	 *	Now perform SGML shortcut processing
	 *
	 *	<> means last tag used
	 *	</> means end of current top tag
	 */
	 
	if(strcmp(tb, "")==0 && tokens_tail)
		tb = tokens_tail->tag;
	if(strcmp(tb, "/")==0 && stack)
	{
		/* Generate a closing tag */
		*buf = '/';
		strncpy(buf+1, stack->tag, sizeof(buf)-2);
		buf[sizeof(buf)-1] = 0;
		tb = buf;
	}
	/*
	 *	Push or pop tags
	 */
		
	if(*tb=='/')
	{
		if(token->attr)
			parser_warning("Attributes on closing tag");
		token_pop_to(token->tag+1);
	}		
	token->data = p;
	token_push(token);
	return p;
}


struct token_stack *parser(char *in)
{	
	static char rootbuf[1] = {'\0'};	/* Must be writable */
	struct token_stack *root = tag_alloc("<DOCUMENT>");
	char *p, *t;
	token_push(root);
	root->data = rootbuf;
	
	/* This is more than needed, we can keep a circular buffer because
	   as we render tags we throw away head data */
	   
	t=p=strdup(in);
	
	while(*p)
	{	
		while(*p && *p!='<')
			p++;
		if(*p == 0)
			break;
		*p++=0;
		p = parse_tag(p);
	}
	
	token_pop_to("<DOCUMENT>");
	free(t);
	return root;
}

void parse_complete(void)
{
	struct token_stack *t;
	
	while(tokens)
	{
		t=tokens;
		tokens=t->next;
		while(t->attr)
		{
			struct token_attr *a = t->attr;
			t->attr=a->next;
			free(a);
		}
		free(t);
	}
	tokens = NULL;
	tokens_tail = NULL;
	stack = NULL;
}

#if 0
int main(int argc, char *argv[])
{
	struct render_state rt;
	char buf[16384];
	int l = read(0, buf, 16384);
	buf[l]=0;	
	render_reset(&rt);
	parser(buf);
	exit(0);
}
#endif

==> tuxurf-0.03.tar.gz: HTW/token.c <==
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "parser.h"

#define html_eof(x)	((*x)==0)

struct token_stack *tokens = NULL;
struct token_stack *tokens_tail = NULL;
struct token_stack *stack = NULL;

static void token_push(struct token_stack *ts)
{
	ts->link = NULL;
	if(tokens == NULL)
		tokens = ts;
	else
		tokens_tail->link = ts;
	tokens_tail = ts;
	ts->next = stack;
	if(stack && stack->done == 0)
	{
		render_text(stack->data);
		stack->done = 1;
	}
	ts->render = find_renderer(ts->tag);
	if(ts->render)
		ts->render->begin(ts);
	stack = ts;
}

static void token_pop_to(char *tag)
{
	char buf[256];
	struct token_stack *ts = stack;

	while(ts!=NULL)
	{
		if(strcasecmp(ts->tag, tag)==0)
			break;
		ts=ts->next;
	}	
	if(ts==NULL)
	{
		snprintf(buf, 256, "No opening tag found for '%s'", tag);
		parser_debug(buf);
		return;
	}
	/*
	 *	Pop entries
	 */
	
	if(!stack->done)
		render_text(stack->data);
	stack->done = 1;
	
	while(stack != ts->next)
	{
		if(stack->render)
			stack->render->end(stack);
		stack = stack->next;
	}
}

/*
 *	Handle implied closures
 */
 
void token_peek_pop(char *tag, char *until)
{
	if(stack && strcasecmp(stack->tag, tag)==0)
	{
		token_pop_to(tag);
		return;
	}
	if(until)
	{
		struct token_stack *p = stack;
		while(p!=NULL)
		{
			/* Found the bounding tag first. No implies */
			if(strcasecmp(p->tag, until)==0)
				return;
			/* Found implied close, pop back to it */
			if(strcasecmp(p->tag, tag)==0)
			{
				token_pop_to(tag);
				return;
			}
			p=p->next;
		}
	}
}
	
static void tag_append(struct token_stack *token, struct token_attr *attr)
{
	struct token_attr **a = &token->attr;
	
	while((*a)!=NULL)
	{
		if(strcasecmp((*a)->key, attr->key)==0)
			parser_debug("Duplicate attribute");
		a=&((*a)->next);
	}
	attr->next = NULL;
	*a = attr;
}

char *tag_attr_default(struct token_stack *token, char *key, char *def)
{
	struct token_attr *a = token->attr;
	
	while(a!=NULL)
	{
		if(strcasecmp(a->key, key)==0)
			return a->val;
		a=a->next;
	}
	return def;
}
	
char *tag_attr(struct token_stack *token, char *key)
{
	return tag_attr_default(token, key, NULL);
}

static struct token_stack *tag_alloc(char *key)
{
	struct token_stack *token = (struct token_stack *)malloc(sizeof(struct token_stack) + strlen(key)+1);
	if(token==NULL)
		parser_fatal("Out of memory");
	token->tag = (char *)(token+1);
	strcpy(token->tag, key);
	token->next = NULL;
	token->link = NULL;
	token->done = 0;
	token->render = NULL;
	token->data = NULL;
	token->rval = 0L;
	token->attr = NULL;
	return token;
}

static struct token_attr *tag_attr_alloc(char * key, char *val)
{
	struct token_attr *attr;
	
	
	attr = (struct token_attr *)malloc(sizeof(struct token_attr) +
						strlen(key)+strlen(val)+2);
	if(attr==NULL)
		parser_fatal("Out of memory");
	attr->key = (char *)(attr+1);
	attr->val = attr->key + strlen(key)+1;
	attr->next = NULL;
	strcpy(attr->key, key);
	strcpy(attr->val, val);
	return attr;
}

static char *parse_quoted_value(char *p, char **v, int *done)
{
	char quote = *p++;

	*v = p;
		
	while(!html_eof(p) && *p!=quote)
	{
		p++;
	}
	if(html_eof(p))
	{
		parser_debug("Missing quote");
		*done=1;
		return p;
	}
	else
		*p++=0;
		
	return p;
}	
	
static char *parse_attr_value(char *p, char **v, int *done)
{
	static char nullstr[]="";
	
	if(html_eof(p) || *p=='>')
	{
		*v=nullstr;
		return p;
	}
	
	while(!html_eof(p) && isspace(*p))
		p++;
		
	if(html_eof(p))
		return p;
		
	if(*p=='\'' || *p=='"')
		return parse_quoted_value(p, v, done);
		
	*v = p;
	while(!html_eof(p) && !isspace(*p) && *p!='>')
		p++;
	if(isspace(*p))
		*p++=0;
	if(*p=='>')
	{
		*p++=0;
		*done=1;
	}
	return p;
}

static char *parse_attr(struct token_stack *token, char *p, int *done)
{
	struct token_attr *attr;
	char *k;
	char *v = "";
	while(!html_eof(p) && isspace(*p))
		p++;
	if(html_eof(p) || *p=='>')
	{
		*done = 1;
		if(*p=='>')
			*p++=0;
		return p;
	}
	k=p;
	while(!html_eof(p) && !isspace(*p) && *p != '>' && *p != '=')
		p++;
	if(*p=='=')
	{
		*p++=0;
		p = parse_attr_value(p, &v, done);
	}
	else if(*p == '>')
	{
		*done = 1;
		*p++=0;
	}
	else if(!html_eof(p))
		*p++=0;
				
	attr = tag_attr_alloc(k, v);
	tag_append(token, attr);
	return p;		
}


static char *parse_tag(char *p)
{
	char *tb = p;
	int done = 0;
	struct token_stack *token;
	char buf[32];
	
	/*
	 *	Check for SGML comment [CDATA]
	 */
	 
	if(memcmp(p, "!--", 3)==0)
	{
		while(!html_eof(p) && strncmp(p, "-->",3))
			p++;
		return p;
	}
	
	while(!html_eof(p))
	{
		if(isspace(*p))
			break;
		if(*p=='>')
			break;
		p++;
	}
	if(html_eof(p))
	{
		parser_debug("Unexpected end of document");
		return p;
	}
	if(*p=='>')
		done = 1;
		
	*p++ = 0;
	
	token = tag_alloc(tb);
	
	while(!html_eof(p) && *p!='>' && !done)
		p=parse_attr(token, p, &done);
	
	if(html_eof(p) && !done)
		parser_debug("Unexpected end of document");
	
	if(*p=='>')
		*p++ = 0;
		
	/*
	 *	Now perform SGML shortcut processing
	 *
	 *	<> means last tag used
	 *	</> means end of current top tag
	 */
	 
	if(strcmp(tb, "")==0 && tokens_tail)
		tb = tokens_tail->tag;
	if(strcmp(tb, "/")==0 && stack)
	{
		/* Generate a closing tag */
		*buf = '/';
		strncpy(buf+1, stack->tag, sizeof(buf)-2);
		buf[sizeof(buf)-1] = 0;
		tb = buf;
	}
	/*
	 *	Push or pop tags
	 */
		
	if(*tb=='/')
	{
		if(token->attr)
			parser_debug("Attributes on closing tag");
		token_pop_to(token->tag+1);
	}		
	token->data = p;
	token_push(token);
	return p;
}


struct token_stack *parser(char *in)
{	
	static char rootbuf[1] = {'\0'};	/* Must be writable */
	struct token_stack *root = tag_alloc("<DOCUMENT>");
	char *p, *t;
	token_push(root);
	root->data = rootbuf;
	
	/* This is more than needed, we can keep a circular buffer because
	   as we render tags we throw away head data */
	   
	t=p=strdup(in);
	
	while(!html_eof(p))
	{	
		while(*p && *p!='<')
			p++;
		if(html_eof(p))
			break;
		*p++=0;
		p = parse_tag(p);
	}
	
	token_pop_to("<DOCUMENT>");
	free(t);
	return root;
}

void parse_complete(void)
{
	struct token_stack *t;
	
	while(tokens)
	{
		t=tokens;
		tokens=t->next;
		while(t->attr)
		{
			struct token_attr *a = t->attr;
			t->attr=a->next;
			free(a);
		}
		free(t);
	}
	tokens = NULL;
	tokens_tail = NULL;
	stack = NULL;
}

#if 0
int main(int argc, char *argv[])
{
	struct render_state rt;
	char buf[16384];
	int l = read(0, buf, 16384);
	buf[l]=0;	
	render_reset(&rt);
	parser(buf);
	exit(0);
}
#endif
